/**************************************************************************
 * Written by : Midhun S
 * Date : 13-09-2019
 *************************************************************************/
#include <TeensyView.h>
#include <PushButton.h>
#include <LineSensor.h>
#include <TeensyViewMenu.h>
#include <NXPMotionSense.h>
#include <Encoder.h>
#include <EEPROM.h>
#include <Motors.h>
#include <ADC.h>

#define PIN_RESET 15
#define PIN_DC    21
#define PIN_CS    20
#define PIN_SCK   13
#define PIN_MOSI  11

#define WHEEL_DIA 3.70
#define DISTANCE_PER_COUNT (PI*WHEEL_DIA/240.00)
#define WHEEL_SPACING 15.00

#define RIGHT 8
#define LEFT  6

#define NORMAL  0
#define ACCL    1

#define OFF     0
#define ON      1
#define TOGGLE  2

#define LEFT_BUTTON   25
#define CENTRE_BUTTON 26
#define RIGHT_BUTTON  27

uint32_t currTime, loopTime, prevTime=0;
uint16_t binaryValue, sensorValue[16];
int linePosition, error, lastError=0, powerDifference;
uint8_t binaryCount, baseSpeed = EEPROM.read(97), maxSpeed = EEPROM.read(97), acclSpeed = EEPROM.read(104), rotateSpeed = EEPROM.read(98);
float Kp = EEPROM.read(99), Kd = EEPROM.read(100);

float leftDistance, rightDistance, totalDistance=0.0, expr, x=0.0, y=0.0, theta=0.0;
int leftTicks, rightTicks, leftTurnCount = 0, rightTurnCount = 0;

TeensyView oled(PIN_RESET, PIN_DC, PIN_CS, PIN_SCK, PIN_MOSI);
PushButton Button(LEFT_BUTTON, CENTRE_BUTTON, RIGHT_BUTTON);
IntervalTimer loopTimer;
TeensyViewMenu menu;
Encoder right(9,27);
Encoder left(4,0);
LineSensor QTRX;
Motors motor;

uint8_t  runMode = EEPROM.read(101), stages = EEPROM.read(105);
uint16_t stageDistance[5][2], k = 0;
uint16_t lapDistance;

/*
uint8_t accSpeed = 200, stage = 1;
uint16_t lapDist = 720;
uint16_t start0 = 110, start1 = 900, start2 = 1500, start3 = 1900, start4 = 2500;
uint16_t end0 = 310, end1 = 1200, end2 = 1800, end3 = 2100, end4 = 2800;
*/

void setup()
{
 /*
  EEPROM.put(102, lapDist);
  EEPROM.write(104, accSpeed);
  EEPROM.write(105, stage);
  EEPROM.put(106, start0);
  EEPROM.put(108, end0);
  EEPROM.put(110, start1);
  EEPROM.put(112, end1);
  EEPROM.put(114, start2);
  EEPROM.put(116, end2);
  EEPROM.put(118, start3);
  EEPROM.put(120, end3);
  EEPROM.put(122, start4);
  EEPROM.put(124, end4);
 */ 
  oled.begin();
  menu.intro();
  
  EEPROM.get(102, lapDistance);
  for(int i=0; i<stages; i++)
    for(int j=0; j<2; j++)
    {
      EEPROM.get(106+k, stageDistance[i][j]);
      k += 2;
    }
  //Configure LED pins to output mode
  pinMode(LEFT, OUTPUT);
  pinMode(RIGHT, OUTPUT);
}

void loop()
{
  currTime = micros();
  loopTime = currTime-prevTime;
  prevTime = currTime;
  /********************* Read line sensors **************************/
  QTRX.getSensorsNormalized(sensorValue, ON);
  binaryValue = QTRX.getSensorsBinary(sensorValue);
  binaryCount = QTRX.countBinary(binaryValue);
  leftTicks = left.read();
  rightTicks = right.read();
  left.write(0);
  right.write(0);
  leftDistance = (float)leftTicks*DISTANCE_PER_COUNT;
  rightDistance = (float)rightTicks*DISTANCE_PER_COUNT;
  totalDistance += (leftDistance+rightDistance)/2;
  
  if(leftTicks == rightTicks)
  {
    x += leftDistance*cos(theta);
    y += rightDistance*sin(theta);
  }
  else
  {
    expr = WHEEL_SPACING*(rightDistance+leftDistance)/2/(rightDistance-leftDistance);
    x += expr*(sin((rightDistance-leftDistance)/WHEEL_SPACING+theta)-sin(theta));
    y -= expr*(cos((rightDistance-leftDistance)/WHEEL_SPACING+theta)-cos(theta));
    theta += (rightDistance-leftDistance)/WHEEL_SPACING;
    if(theta > PI)
      theta -= (2*PI);
    else if(theta < -PI)
      theta += (2*PI);
  }
  oled.clear(PAGE);
  oled.setCursor(0,0);
  /*
  oled.clear(PAGE);
  for(int i=0; i<16; i++)
  {
      oled.setCursor((15-i)*6,0);
      if(binaryValue & (1<<i))
        oled.print("1");
      else
        oled.print("0");
  }
  oled.display();
*/
  switch(binaryCount)
  {
    case 0  : error = lastError = 0;  break;
    case 1  : PID();  break;
    case 2  : if(binaryValue%3 == 0)
                PID();
              break;
    case 3  : if(binaryValue%7 == 0)
                PID();
              break;
    
    case 4  : if(binaryValue%15 == 0)
                PID();
              else if ((binaryValue == 0b1100000110000000) || (binaryValue == 0b1000001110000000) || (binaryValue == 0b1000000111000000))
                rotate(LEFT);
              else if((binaryValue == 0b0000000110000011) || (binaryValue == 0b0000001110000001) || (binaryValue == 0b0000000111000001))
                rotate(RIGHT);
              break;
              
    case 5  : if((binaryValue == 0b1100001110000000) || (binaryValue == 0b1100000111000000) || (binaryValue == 0b1000001111000000))
                rotate(LEFT);
              else if((binaryValue == 0b0000001110000011) || (binaryValue == 0b0000000111000011) || (binaryValue == 0b0000001111000001))
                rotate(RIGHT);
              break;
              
    case 6  :               
    case 7  : if((binaryValue & 0b0000000000111111) == 0)
                rotate(LEFT);
              else if((binaryValue & 0b1111110000000000) == 0)
                rotate(RIGHT);
              break;
                            
    case 8  :               
    case 9  : if((binaryValue & 0b0000000000011111) == 0)
                rotate(LEFT);
              else if((binaryValue & 0b1111100000000000) == 0)
                rotate(RIGHT);
              break;
              
    case 10 :               
    case 11 : if((binaryValue & 0b0000000000001111) == 0)
                rotate(LEFT);
              else if((binaryValue & 0b1111000000000000) == 0)
                rotate(RIGHT);
              break;
              
    case 12 : 
    case 13 : if((binaryValue & 0b0000000000000111) == 0)
                rotate(LEFT);
              else if((binaryValue & 0b1110000000000000) == 0)
                rotate(RIGHT);
              break;
              
    case 14 : if((binaryValue & 0b0000000000000011) == 0)
                rotate(LEFT);
              else if((binaryValue & 0b1100000000000000) == 0)
                rotate(RIGHT);
              break;
              
    case 15 : 
    case 16 : /*Code to turn left at a T-section
              if((totalDistance > 750) && (leftTurn < 5))
              {
                do
                { 
                  QTRX.getSensorsNormalized(sensorValue, TOGGLE);
                  binaryValue = QTRX.getSensorsBinary(sensorValue);
                  // Skip turn if distance covered while within loop is > 30cm
                  if((totalDistance - loopStartDistance) > 30)
                  {
                    digitalWriteFast(side, LOW);
                    return;
                  }
                }while(binaryValue & 0b0000000110000000);
              */
              rotate(LEFT);
              lastError = 0;
              break;
  }

  if(runMode == ACCL)
  {
    switch(stages)
    {
      case 5 :  if((totalDistance > stageDistance[0][0]) && (totalDistance < stageDistance[0][1]))
                  maxSpeed = acclSpeed;
                else if((totalDistance > stageDistance[1][0]) && (totalDistance < stageDistance[1][1]))
                  maxSpeed = acclSpeed;
                else if((totalDistance > stageDistance[2][0]) && (totalDistance < stageDistance[2][1]))
                  maxSpeed = acclSpeed;
                else if((totalDistance > stageDistance[3][0]) && (totalDistance < stageDistance[3][1]))
                  maxSpeed = acclSpeed;
                else if((totalDistance > stageDistance[4][0]) && (totalDistance < stageDistance[4][1]))
                  maxSpeed = acclSpeed;
                else
                  maxSpeed = baseSpeed;
                break;

      case 4 :  if((totalDistance > stageDistance[0][0]) && (totalDistance < stageDistance[0][1]))
                  maxSpeed = acclSpeed;
                else if((totalDistance > stageDistance[1][0]) && (totalDistance < stageDistance[1][1]))
                  maxSpeed = acclSpeed;
                else if((totalDistance > stageDistance[2][0]) && (totalDistance < stageDistance[2][1]))
                  maxSpeed = acclSpeed;
                else if((totalDistance > stageDistance[3][0]) && (totalDistance < stageDistance[3][1]))
                  maxSpeed = acclSpeed;
                else
                  maxSpeed = baseSpeed;
                break;
                  
      case 3 :  if((totalDistance > stageDistance[0][0]) && (totalDistance < stageDistance[0][1]))
                  maxSpeed = acclSpeed;
                else if((totalDistance > stageDistance[1][0]) && (totalDistance < stageDistance[1][1]))
                  maxSpeed = acclSpeed;
                else if((totalDistance > stageDistance[2][0]) && (totalDistance < stageDistance[2][1]))
                  maxSpeed = acclSpeed;
                else
                  maxSpeed = baseSpeed;
                break;

      case 2 :  if((totalDistance > stageDistance[0][0]) && (totalDistance < stageDistance[0][1]))
                  maxSpeed = acclSpeed;
                else if((totalDistance > stageDistance[1][0]) && (totalDistance < stageDistance[1][1]))
                  maxSpeed = acclSpeed;
                else
                  maxSpeed = baseSpeed;
                break;
                
      case 1 :  if((totalDistance > stageDistance[0][0]) && (totalDistance < stageDistance[0][1]))
                  maxSpeed = acclSpeed;
                else
                  maxSpeed = baseSpeed;
                break;
    }
    if((totalDistance > lapDistance) && (((int)(theta*180/PI)%360) < 10))
    {      
      motor.setSpeed(0,0);
      oled.clear(PAGE);
      oled.setCursor(0,0);
      oled.print("Distance = ");
      oled.print((int)totalDistance);
      oled.setCursor(0,10);
      oled.print("Theta = ");
      oled.print((int)(theta*180/PI));
      oled.display();
      while(1);
    }
  /*else if((totalDistance > 500) && (leftCount < 2))  //will be mostly true if > is used
    {      
    }
    else if((totalDistance > 1000) && (leftCount < 7))
    {      
    }
   */
  }
  oled.print(totalDistance);
  oled.setCursor(0,10);
  oled.print(leftTurnCount);
  oled.print(",");
  oled.print(rightTurnCount);
  oled.display();
}
/************************************** PID ROUTINE *****************************************/
void PID()
{
  error = QTRX.getLinePosition(sensorValue);
  powerDifference = (Kp/1000)*(float)error + (Kd/1000)*(float)(error-lastError);
  lastError = error;
  powerDifference = constrain(powerDifference, -2*maxSpeed, 2*maxSpeed);
  if(powerDifference < 0)
    motor.setSpeed(maxSpeed+powerDifference, maxSpeed);
  else
    motor.setSpeed(maxSpeed, maxSpeed-powerDifference);
}
/************************************* ROTATE ROUTINE ****************************************/
void rotate(uint8_t side)
{
  int loopStartDistance = totalDistance;  
  lastError=0;
  digitalWriteFast(side, HIGH);
  //Continue previous action until two sensors at centre is out of the line
  do
  { 
    QTRX.getSensorsNormalized(sensorValue, TOGGLE);
    binaryValue = QTRX.getSensorsBinary(sensorValue);
    
    leftTicks = left.read();
    rightTicks = right.read();
    left.write(0);
    right.write(0);
    leftDistance = (float)leftTicks*DISTANCE_PER_COUNT;
    rightDistance = (float)rightTicks*DISTANCE_PER_COUNT;
    totalDistance += (leftDistance+rightDistance)/2;
  
/*  if(((side == RIGHT) && (binaryValue & 0b1111000000000000)) || ((side == LEFT) && (binaryValue & 0b0000000000001111)))
    {
      digitalWriteFast(side, LOW);
      return;
    }
 */ 
    // Skip turn if distance covered while within loop is > 30cm
    if((totalDistance - loopStartDistance) > 30)
    {
      digitalWriteFast(side, LOW);
      return;
    }
  }while(binaryValue & 0b0000000110000000); //May have to check middle 4 sensors if there is loop in track
  
  //Perform rotation
  if(side == RIGHT)
  {
    motor.setSpeed(rotateSpeed, -rotateSpeed);
    rightTurnCount++;
  }
  else if(side == LEFT)
  {
    motor.setSpeed(-rotateSpeed, rotateSpeed);
    leftTurnCount++;
  }
  digitalWriteFast(side, LOW);
}
